import os
import re
import sys
import timeit
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
import psutil
from mpi4py import MPI
from scipy.fft import fft, fftfreq, fftshift
from scipy.integrate import trapezoid
from scipy.signal.windows import hann
from tqdm import tqdm

from .h5 import h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_turb_cospectrum_xpln(self, **kwargs):
    '''
    Calculate 1D FFT cospectrum in [z] and [t] at every [x,y]
    - Designed for analyzing unsteady, thin planes in [x]
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'rgd.calc_turb_cospectrum_xpln()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## Assert that the opened RGD has fsubtype 'unsteady' (i.e. is NOT a prime file)
    if (self.fsubtype!='unsteady'):
        raise ValueError
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    if not h5py.h5.get_config().mpi:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    rx = kwargs.get('rx',1)
    ry = kwargs.get('ry',1)
    rz = kwargs.get('rz',1)
    rt = kwargs.get('rt',1)
    
    sy = kwargs.get('sy',1) ## number of [y] layers to read at a time
    if not isinstance(sy,int) or (sy<1):
        raise TypeError('sy should be a positive non-zero int')
    
    n_threads = kwargs.get('n_threads',1)
    fn_h5_out = kwargs.get('fn_h5_out',None) ## Filename for output HDF5 (.h5) file (default chosen later)
    n_win     = kwargs.get('n_win',16)       ## Number of segment windows for [t] PSD calc
    
    ## Debug Rank:Proc Affinity
    #pp = psutil.Process()
    #print(f"[Rank {self.rank}] sees CPUs: {pp.cpu_affinity()}  |  n_threads={n_threads}  |  OMP_NUM_THREADS={os.environ.get('OMP_NUM_THREADS')}")
    
    ## Only distribute data across [y]
    if (rx!=1):
        raise ValueError('rx!=1')
    if (rz!=1):
        raise ValueError('rz!=1')
    if (rt!=1):
        raise ValueError('rt!=1')
    
    if not isinstance(ry,int) or (ry<1):
        raise ValueError('ry should be a positive non-zero int')
    
    ## Check the choice of ranks per dimension
    if (rx*ry*rz*rt != self.n_ranks):
        raise ValueError('rx*ry*rz*rt != self.n_ranks')
    if (rx>self.nx):
        raise ValueError('rx>self.nx')
    if (ry>self.ny):
        raise ValueError('ry>self.ny')
    if (rz>self.nz):
        raise ValueError('rz>self.nz')
    if (rt>self.nt):
        raise ValueError('rt>self.nt')
    
    if (self.ny%ry!=0):
        raise ValueError('ny not divisible by ry')
    
    ## Distribute 4D data over ranks --> here only in [y]
    ryl_ = np.array_split(np.arange(self.ny,dtype=np.int64),ry)
    ryl = [[b[0],b[-1]+1] for b in ryl_ ]
    ry1,ry2 = ryl[self.rank]
    nyr = ry2 - ry1
    
    ## Check all [y] ranges have same size (current limitation)
    for ryl_ in ryl:
        if not (ryl_[1]-ryl_[0]==nyr):
            raise ValueError('[y] chunks are not even in size')
    
    if (nyr%sy!=0):
        raise ValueError('nyr not divisible by sy')
    
    ## Output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## Automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fname_fft_h5_base = fname_root+'_fft.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fname_fft_h5_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'     , self.fname )
    if verbose: even_print( 'fn_h5_out' , fn_h5_out  )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    ## Data dictionary to be written to .h5 later
    data = {}
    
    ## infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('nx',f'{self.nx:d}')
    if verbose: even_print('ny',f'{self.ny:d}')
    if verbose: even_print('nz',f'{self.nz:d}')
    if verbose: even_print('nt',f'{self.nt:d}')
    if verbose: even_print('ngp',f'{self.ngp/1e6:0.1f} [M]')
    if verbose: even_print('sy',f'{sy:d}')
    if verbose: even_print('n_ranks',f'{self.n_ranks:d}')
    if verbose: even_print('n_threads',f'{n_threads:d}')
    if verbose: print(72*'-')
    
    ## 0D freestream scalars
    lchar   = self.lchar   ; data['lchar']   = lchar
    U_inf   = self.U_inf   ; data['U_inf']   = U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = T_inf
    
    #data['M_inf'] = self.M_inf
    data['Ma'] = self.Ma
    data['Pr'] = self.Pr
    
    ## read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    y = np.copy( self['dims/y'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    
    nx = self.nx ; data['nx'] = nx
    ny = self.ny ; data['ny'] = ny
    nz = self.nz ; data['nz'] = nz
    nt = self.nt ; data['nt'] = nt
    
    ## Assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    dz = np.diff(z)[0]
    
    ## Dimensional time vector [s]
    dt = self.dt * self.tchar
    np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    
    ## Measurement time [s]
    t_meas = self.duration * self.tchar
    np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    
    zrange = z.max() - z.min()
    
    data['x'] = x
    data['y'] = y
    data['z'] = z
    
    data['t']      = t
    data['t_meas'] = t_meas
    data['dt']     = dt
    data['dz']     = dz
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## Report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        print(72*'-')
    
    ## Establish [t] windowing
    ## - 50% overlap is hardcoded
    nperseg  = int(2*nt/(n_win+1))          ## nt window for 50% overlap (nominal)
    nperseg  = (nperseg//2)*2               ## make divisible by 2
    noverlap = nperseg//2                   ## nt overlap
    ntu      = nperseg + (n_win-1)*noverlap ## nt used
    overlap_fac = noverlap / nperseg
    
    ## Window for [t] FFT : Hann
    win  = hann(nperseg, sym=False).astype(np.float64) ## Hann window
    win2 = np.mean(win*win)                            ## Window power correction
    
    ## Temporal [t] frequency (f) vector --> Using windows (Welch-type)
    freq_full = fftshift(fftfreq(n=nperseg, d=dt))
    fp        = np.where(freq_full>0)[0] ## Indices of positive values
    freq      = np.copy(freq_full[fp])
    df        = float(freq[1] - freq[0])
    nf        = int(freq.shape[0])
    
    data['n_win']    = n_win
    data['nperseg']  = nperseg
    data['noverlap'] = noverlap
    data['freq']     = freq
    data['df']       = df
    data['nf']       = nf
    
    if verbose:
        even_print('n_win'          , f'{n_win:d}'          )
        even_print('nperseg'        , f'{nperseg:d}'        )
        even_print('noverlap'       , f'{noverlap:d}'       )
        even_print('overlap_fac'    , f'{overlap_fac:0.5f}' )
        even_print('timesteps used' , f'{ntu}/{nt}'         )
        print(72*'-')
    
    if verbose:
        even_print('freq min',f'{freq.min():0.1f} [Hz]')
        even_print('freq max',f'{freq.max():0.1f} [Hz]')
        even_print('df',f'{df:0.1f} [Hz]')
        even_print('nf',f'{nf:d}')
        print(72*'-')
    
    ## Spatial [z] wavenumber (kz) vector
    kz_full = fftshift(fftfreq(n=nz, d=dz)) * ( 2 * np.pi )
    kzp     = np.where(kz_full>0)[0] ## Indices of positive values
    kz      = np.copy(kz_full[kzp])
    dkz     = float(kz[1] - kz[0])
    nkz     = int(kz.shape[0])
    
    data['kz']  = kz
    data['dkz'] = dkz
    data['nkz'] = nkz
    
    ## λz = 2·π/kz
    lz  = np.copy( 2 * np.pi / kz )
    dlz = float(lz[1] - lz[0])
    nlz = int(lz.shape[0])
    
    data['lz']  = lz
    data['dlz'] = dlz
    data['nlz'] = nlz
    
    if verbose:
        even_print('kz min',f'{kz.min():0.1f} [1/m]')
        even_print('kz max',f'{kz.max():0.1f} [1/m]')
        even_print('dkz',f'{dkz:0.1f} [1/m]')
        even_print('nkz',f'{nkz:d}')
        print(72*'-')
    
    # ===
    
    ## Cospectrum pairs
    ## [ (str:var1, bool:ρ_weighting) , (str:var2, bool:ρ_weighting) ]
    fft_combis = [
    
    [ ('u',True) , ('v',True) ], ## [ ρ·u″ , ρ·v″ ]
    [ ('u',True) , ('u',True) ], ## [ ρ·u″ , ρ·u″ ]
    [ ('v',True) , ('v',True) ], ## [ ρ·v″ , ρ·v″ ]
    [ ('w',True) , ('w',True) ], ## [ ρ·w″ , ρ·w″ ]
    
    [ ('u',False) , ('v',False) ], ## [ u′ , v′ ]
    [ ('u',False) , ('u',False) ], ## [ u′ , u′ ]
    [ ('v',False) , ('v',False) ], ## [ v′ , v′ ]
    [ ('w',False) , ('w',False) ], ## [ w′ , w′ ]
    
    [ ('p',False) , ('u',False) ], ## [ p′ , u′ ]
    [ ('p',False) , ('v',False) ], ## [ p′ , v′ ]
    [ ('p',False) , ('w',False) ], ## [ p′ , w′ ]
    
    [ ('p',False) , ('u',True) ], ## [ p′ , ρ·u″ ]
    [ ('p',False) , ('v',True) ], ## [ p′ , ρ·v″ ]
    [ ('p',False) , ('w',True) ], ## [ p′ , ρ·w″ ]
    
    [ ('p',False)   , ('p',False)   ], ## [ p′ , p′ ]
    [ ('T',False)   , ('T',False)   ], ## [ T′ , T′ ]
    [ ('rho',False) , ('rho',False) ], ## [ ρ′ , ρ′ ]
    
    ]
    
    ## Generate FFT cospectrum scalar names
    scalars = []
    for fft_combi in fft_combis:
        if not isinstance(fft_combi,list):
            raise RuntimeError
        if not len(fft_combi)==2:
            raise RuntimeError
        
        sL = fft_combi[0][0]
        do_density_weighting_L = fft_combi[0][1]
        if do_density_weighting_L:
            sLs = f'r{sL}II'
        else:
            sLs = f'{sL}I'
        
        sR = fft_combi[1][0]
        do_density_weighting_R = fft_combi[1][1]
        if do_density_weighting_R:
            sRs = f'r{sR}II'
        else:
            sRs = f'{sR}I'
        
        scalars.append(f'{sLs}_{sRs}')
    
    ## Generate avg scalar names
    scalars_Re_avg = []
    scalars_Fv_avg = []
    for fft_combi in fft_combis:
        sL = fft_combi[0][0]
        do_density_weighting_L = fft_combi[0][1]
        sR = fft_combi[1][0]
        do_density_weighting_R = fft_combi[1][1]
        
        if do_density_weighting_L or do_density_weighting_R:
            if 'rho' not in scalars_Re_avg:
                scalars_Re_avg.append('rho')
        
        if do_density_weighting_L:
            if sL not in scalars_Fv_avg:
                scalars_Fv_avg.append(sL)
        else:
            if sL not in scalars_Re_avg:
                scalars_Re_avg.append(sL)
        
        if do_density_weighting_R:
            if sR not in scalars_Fv_avg:
                scalars_Fv_avg.append(sR)
        else:
            if sR not in scalars_Re_avg:
                scalars_Re_avg.append(sR)
    
    ## numpy formatted arrays: buffers for PSD & other data (rank-local)
    Pkz        = np.zeros(shape=(nyr,nkz ) , dtype={'names':scalars        , 'formats':[ np.dtype(np.float64) for s in scalars        ] })
    Pf         = np.zeros(shape=(nyr,nf  ) , dtype={'names':scalars        , 'formats':[ np.dtype(np.float64) for s in scalars        ] })
    covariance = np.zeros(shape=(nyr,    ) , dtype={'names':scalars        , 'formats':[ np.dtype(np.float64) for s in scalars        ] })
    avg_Re     = np.zeros(shape=(nyr,    ) , dtype={'names':scalars_Re_avg , 'formats':[ np.dtype(np.float64) for s in scalars_Re_avg ] })
    avg_Fv     = np.zeros(shape=(nyr,    ) , dtype={'names':scalars_Fv_avg , 'formats':[ np.dtype(np.float64) for s in scalars_Fv_avg ] })
    
    if verbose:
        even_print('n turb spectrum scalar combinations' , '%i'%(len(fft_combis),))
        print(72*'-')
    
    # ==============================================================
    # Check memory
    # ==============================================================
    
    hostname = MPI.Get_processor_name()
    mem_free_gb = psutil.virtual_memory().free / 1024**3
    G = self.comm.gather([ self.rank , hostname , mem_free_gb ], root=0)
    G = self.comm.bcast(G, root=0)
    
    host_mem = {}
    for rank, host, mem in G:
        if host not in host_mem or mem < host_mem[host]:
            host_mem[host] = mem
    total_free = sum(host_mem.values())
    
    if verbose:
        #print(72*'-')
        for key,value in host_mem.items():
            even_print(f'RAM free {key}', f'{int(np.floor(value)):d} [GB]')
        even_print('RAM free (local,min)', f'{int(np.floor(min(host_mem.values()))):d} [GB]')
        even_print('RAM free (global)', f'{int(np.floor(total_free)):d} [GB]')
    
    shape_read = (nx,sy,nz,nt) ## local
    if verbose: even_print('read shape (local)', f'[{nx:d},{sy:d},{nz:d},{nt:d}]')
    data_gb = np.dtype(np.float64).itemsize * np.prod(shape_read) / 1024**3
    if verbose: even_print('read size (global)', f'{int(np.ceil(data_gb*ry)):d} [GB]')
    
    if verbose: even_print('read size (global) ×8', f'{int(np.ceil(data_gb*ry*8)):d} [GB]')
    ram_usage_est = data_gb*ry*8/total_free
    if verbose: even_print('RAM usage estimate', f'{100*ram_usage_est:0.1f} [%]')
    
    self.comm.Barrier()
    if (ram_usage_est>0.80):
        print('RAM consumption might be too high. exiting.')
        self.comm.Abort(1)
    
    # ==============================================================
    # Main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            #total=len(fft_combis)*cy,
            total=len(fft_combis)*(nyr//sy),
            ncols=100,
            desc='fft',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#00ffff',
            )
    
    for cci,cc in enumerate(fft_combis): ## FFT pairs
        
        if verbose: tqdm.write(72*'-')
        
        scalar_L = cc[0][0]
        do_density_weighting_L = cc[0][1]
        scalar_R = cc[1][0]
        do_density_weighting_R = cc[1][1]
        
        if do_density_weighting_L:
            sLs  = f'r{scalar_L}II'
            sLsF = f'ρ·{scalar_L}″'
        else:
            sLs  = f'{scalar_L}I'
            sLsF = f'{scalar_L}′'
        
        if do_density_weighting_R:
            sRs  = f'r{scalar_R}II'
            sRsF = f'ρ·{scalar_R}″'
        else:
            sRs  = f'{scalar_R}I'
            sRsF = f'{scalar_R}′'
        
        msg = f'[{sLsF},{sRsF}]'
        
        dset_L   = self[f'data/{scalar_L}']
        dset_R   = self[f'data/{scalar_R}']
        dset_rho = self['data/rho']
        
        scalar = scalars[cci]
        
        ## Assert scalar name
        if (f'{sLs}_{sRs}' != scalar):
            raise RuntimeError(f"'{sLs}_{sRs}' != '{scalar}'")
        
        ## [y] loop outer (grid subchunks within rank)
        for ci in range(nyr//sy):
            
            ## Buffers
            data_L = np.zeros((nx,sy,nz,nt), dtype=np.float64)
            data_R = np.zeros((nx,sy,nz,nt), dtype=np.float64)
            if do_density_weighting_L or do_density_weighting_R:
                rho = np.zeros((nx,sy,nz,nt), dtype=np.float64)
            else:
                rho = None
            
            cy1 = ry1 + ci*sy
            cy2 = cy1 + sy
            nyc = cy2 - cy1
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## Read data L
            n_scalars_read = 1 ## initialize
            scalar_str = scalar_L
            with dset_L.collective:
                data_L[:,:,:,:] = dset_L[:,:,cy1:cy2,:].T
            
            ## Read data R (if != data L)
            if (scalar_L==scalar_R):
                data_R[:,:,:,:] = data_L[:,:,:,:]
            else:
                n_scalars_read += 1
                scalar_str += f',{scalar_R}'
                with dset_R.collective:
                    data_R[:,:,:,:] = dset_R[:,:,cy1:cy2,:].T
            
            ## Read ρ
            if do_density_weighting_L or do_density_weighting_R:
                n_scalars_read += 1
                scalar_str += ',ρ'
                with dset_rho.collective:
                    rho[:,:,:,:] = dset_rho[:,:,cy1:cy2,:].T
            else:
                rho = None
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = n_scalars_read * ( nx * ry * (cy2-cy1) * nz * nt * dset_L.dtype.itemsize ) / 1024**3
            if verbose:
                tqdm.write(even_print(f'read: {scalar_str}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## data_L and data_R should be [nx,nyc,nz,nt] where nyc is the chunk [y] range
            if ( data_L.shape != (nx,nyc,nz,nt) ) or ( data_R.shape != (nx,nyc,nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            if (rho is not None) and ( rho.shape != (nx,nyc,nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === Redimensionalize unsteady data
            
            if scalar_L in ['u','v','w',]:
                data_L *= U_inf
            elif scalar_L in ['p',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['T',]:
                data_L *= T_inf
            elif scalar_L in ['rho',]:
                data_L *= rho_inf
            else:
                raise RuntimeError
            
            if scalar_R in ['u','v','w',]:
                data_R *= U_inf
            elif scalar_R in ['p',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['T',]:
                data_R *= T_inf
            elif scalar_R in ['rho',]:
                data_R *= rho_inf
            else:
                raise RuntimeError
            
            if (rho is not None):
                rho *= rho_inf
            
            # === Compute mean-removed data
            
            ## Reynolds avg(□) or Favre avg(ρ·□)/avg(ρ) in [t]
            if do_density_weighting_L:
                rho_avg     = np.mean(        rho , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                data_L_avg  = np.mean( rho*data_L , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                data_L_avg /= rho_avg
            else:
                data_L_avg = np.mean( data_L , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
            
            ## Reynolds avg(□) or Favre avg(ρ·□)/avg(ρ) in [t]
            if do_density_weighting_R:
                rho_avg     = np.mean(        rho , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                data_R_avg  = np.mean( rho*data_R , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                data_R_avg /= rho_avg
            else:
                data_R_avg = np.mean( data_R , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
            
            ## Reynolds prime □′ or Favre prime □″
            data_L -= data_L_avg
            data_R -= data_R_avg
            
            ## LEFT
            ## Assert stationarity + definition Re/Fv averaging
            ## avg(□′)==0 or avg(ρ·□″)==0
            if do_density_weighting_L:
                a_ = np.mean(rho*data_L, axis=3, dtype=np.float64, keepdims=True)
            else:
                a_ = np.mean(data_L, axis=3, dtype=np.float64, keepdims=True)
            if not np.allclose( a_, np.zeros_like(a_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            ## RIGHT
            ## Assert stationarity + definition Re/Fv averaging
            ## avg(□′)==0 or avg(ρ·□″)==0
            if do_density_weighting_R:
                a_ = np.mean(rho*data_R, axis=3, dtype=np.float64, keepdims=True)
            else:
                a_ = np.mean(data_R, axis=3, dtype=np.float64, keepdims=True)
            if not np.allclose( a_, np.zeros_like(a_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            ## Covariance: <□′·□′> OR <ρ□″·ρ□″> etc. --> note that this is NOT the typical Favre <ρ·□″□″> except in special cases
            if do_density_weighting_L and do_density_weighting_R:
                covariance_ = np.mean( rho*data_L * rho*data_R , axis=3 , dtype=np.float64, keepdims=True)
            elif do_density_weighting_L and not do_density_weighting_R:
                covariance_ = np.mean( rho*data_L * data_R , axis=3 , dtype=np.float64, keepdims=True)
            elif not do_density_weighting_L and do_density_weighting_R:
                covariance_ = np.mean( data_L * rho*data_R , axis=3 , dtype=np.float64, keepdims=True)
            elif not do_density_weighting_L and not do_density_weighting_R:
                covariance_ = np.mean( data_L * data_R , axis=3 , dtype=np.float64, keepdims=True)
            else:
                raise RuntimeError
            
            ## Write this chunk/scalar's covariance to covariance buffer
            ## avg over [x,z] : [x,y,z,1] --> [y]
            yiA = cy1 - ry1
            yiB = cy2 - ry1
            covariance[scalar][yiA:yiB] = np.squeeze( np.mean( covariance_ , axis=(0,2,3) , dtype=np.float64) )
            
            ## Write (rank-local) 1D [y] averages
            if do_density_weighting_L or do_density_weighting_R:
                avg_Re['rho'][yiA:yiB] = np.squeeze( np.mean( rho_avg    , axis=(0,2,3) , dtype=np.float64) )
            
            if do_density_weighting_L:
                avg_Fv[scalar_L][yiA:yiB] = np.squeeze( np.mean( data_L_avg , axis=(0,2,3) , dtype=np.float64) )
            else:
                avg_Re[scalar_L][yiA:yiB] = np.squeeze( np.mean( data_L_avg , axis=(0,2,3) , dtype=np.float64) )
            
            if do_density_weighting_R:
                avg_Fv[scalar_R][yiA:yiB] = np.squeeze( np.mean( data_R_avg , axis=(0,2,3) , dtype=np.float64) )
            else:
                avg_Re[scalar_R][yiA:yiB] = np.squeeze( np.mean( data_R_avg , axis=(0,2,3) , dtype=np.float64) )
            
            # ===============================================================================
            # At this point you have 4D [x,y,z,t] □′ or □″ data
            # ===============================================================================
            
            ## [y] loop inner ([y] indices within subdivision within rank)
            for yi in range(cy1,cy2):
                
                yii  = yi - cy1 ## chunk local
                yiii = yi - ry1 ## rank local
                
                # ===========================================================================
                # FFT in [t]
                # ===========================================================================
                
                ## Accumulator for multiple time windows
                P = np.full((nperseg,), 0., dtype=np.float64)
                
                ## [t] Window loop
                for wi in range(n_win):
                    tiA = wi*noverlap
                    tiB = tiA + nperseg
                    if tiB > nt:
                        raise ValueError
                    
                    ## [x,z,t]
                    uL = np.full((nx,nz,nperseg), 0., dtype=np.float64)
                    uR = np.full((nx,nz,nperseg), 0., dtype=np.float64)
                    
                    ## 3D (x,z,t) □′ or ρ·□″ arrays
                    if do_density_weighting_L:
                        uL[:,:,:] = rho[:,yii,:,tiA:tiB] * data_L[:,yii,:,tiA:tiB]
                    else:
                        uL[:,:,:] = data_L[:,yii,:,tiA:tiB]
                    
                    if do_density_weighting_R:
                        uR[:,:,:] = rho[:,yii,:,tiA:tiB] * data_R[:,yii,:,tiA:tiB]
                    else:
                        uR[:,:,:] = data_R[:,yii,:,tiA:tiB]
                    
                    ## Apply Hann window to [t] dimension
                    uL *= win[np.newaxis,np.newaxis,:]
                    uR *= win[np.newaxis,np.newaxis,:]
                    
                    Ai   = fftshift( fft(uL, axis=2, workers=n_threads), axes=2) ## [t] FFT
                    Aj   = fftshift( fft(uR, axis=2, workers=n_threads), axes=2) ## [t] FFT
                    Pij  = np.real( Ai * Aj.conj() )
                    Pseg = np.mean(Pij, axis=(0,1)) ## [x,z] average
                    
                    ## Convert to density
                    Pseg /= nperseg**2 ## nperseg is nt per window
                    Pseg /= df
                    
                    ## Compensate window power loss
                    Pseg /= win2
                    
                    P += (1/n_win)*Pseg ## Add this window's contribution to accumulator
                
                ## Take integral over frequency
                ii = np.argsort(freq_full)
                integral = trapezoid(P[ii], x=freq_full[ii], axis=0)
                
                ## 0D covariance for this [y]
                covar0D = np.mean(covariance_[:,yii,:], dtype=np.float64)
                
                ## !! For freestream locations, the signal is very weak, so the error can be O(5%)
                if False: ## DEBUG for integral / covariance regularization
                    rtol = np.abs(integral-covar0D)/max(np.abs(max(integral,covar0D)),1e-12)
                    if not np.isclose(covar0D, integral, rtol=0.005, atol=1e-5):
                        print(f'j={yi:d} FAIL : rtol={rtol:0.3e} covar0D={covar0D:0.3e} integral={integral:0.3e}')
                    else:
                        print(f'j={yi:d} PASS : rtol={rtol:0.3e} covar0D={covar0D:0.3e} integral={integral:0.3e}')
                
                ## Integral / covariance regularization
                ## Scale by tiny factor to reinstate Parseval
                if np.abs(integral)>0.:
                    fac = covar0D / integral
                    P *= fac
                
                ## Write only freq>0 (for plotting etc) to output buffer
                ## Integrating the output spectrum will yield covar/2
                Pf[scalar][yiii,:] = P[fp]
                
                # ===========================================================================
                # FFT in [z]
                # ===========================================================================
                
                ## [x,z,t]
                uL = np.full((nx,nz,nt), 0., dtype=np.float64)
                uR = np.full((nx,nz,nt), 0., dtype=np.float64)
                
                ## 3D (x,z,t) □′ or ρ·□″ arrays
                if do_density_weighting_L:
                    uL[:,:,:] = rho[:,yii,:,:] * data_L[:,yii,:,:]
                else:
                    uL[:,:,:] = data_L[:,yii,:,:]
                
                if do_density_weighting_R:
                    uR[:,:,:] = rho[:,yii,:,:] * data_R[:,yii,:,:]
                else:
                    uR[:,:,:] = data_R[:,yii,:,:]
                
                Ai  = fftshift( fft(uL, axis=1, workers=n_threads), axes=1) ## [z] FFT
                Aj  = fftshift( fft(uR, axis=1, workers=n_threads), axes=1) ## [z] FFT
                Pij = np.real( Ai * Aj.conj() )
                P   = np.mean(Pij, axis=(0,2)) ## [x,t] average
                
                ## Convert to density
                P /= nz**2
                P /= dkz
                
                ## Take integral over kz
                ii = np.argsort(kz_full)
                integral = trapezoid(P[ii], x=kz_full[ii], axis=0)
                
                ## 0D covariance for this [y]
                covar0D = np.mean(covariance_[:,yii,:], dtype=np.float64)
                
                ## !! For freestream locations, the signal is very weak, so the error can be O(5%)
                if False: ## DEBUG for integral / covariance regularization
                    rtol = np.abs(integral-covar0D)/max(np.abs(max(integral,covar0D)),1e-12)
                    if not np.isclose(covar0D, integral, rtol=0.005, atol=1e-5):
                        print(f'j={yi:d} FAIL : rtol={rtol:0.3e} covar0D={covar0D:0.3e} integral={integral:0.3e}')
                    else:
                        print(f'j={yi:d} PASS : rtol={rtol:0.3e} covar0D={covar0D:0.3e} integral={integral:0.3e}')
                
                ## Integral / covariance regularization
                ## Scale by tiny factor to reinstate Parseval
                if np.abs(integral)>0.:
                    fac = covar0D / integral
                    P *= fac
                
                ## Write only kz>0 (for plotting etc) to output buffer
                ## Integrating the output spectrum will yield covar/2
                Pkz[scalar][yiii,:] = P[kzp]
                
                # ===========================================================================
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            if verbose: tqdm.write(even_print(msg, format_time_string(t_delta), s=True))
            if verbose: progress_bar.update() ## (scalar, [y] chunk) progress
            
            #break ## DEBUG : break y loop
        #break ## DEBUG : break scalar loop
    
    if verbose: progress_bar.close()
    self.comm.Barrier()
    if verbose: print(72*'-')
    
    # ==============================================================
    # Write HDF5 (.h5) file
    # ==============================================================
    
    ## Overwrite outfile!
    ## Open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## Write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## Write numpy arrays
            hfw.create_dataset( 'dims/x'    , data=x    ) ## [m]
            hfw.create_dataset( 'dims/y'    , data=y    ) ## [m]
            hfw.create_dataset( 'dims/z'    , data=z    ) ## [m]
            hfw.create_dataset( 'dims/t'    , data=t    ) ## [s]
            hfw.create_dataset( 'dims/freq' , data=freq ) ## [1/s] | [Hz]
            hfw.create_dataset( 'dims/kz'   , data=kz   ) ## [1/m]
            hfw.create_dataset( 'dims/lz'   , data=lz   ) ## [m]
            
            ## Initialize datasets
            for scalar in scalars:
                hfw.create_dataset( f'covariance/{scalar}' , shape=(ny,)    , dtype=np.float64 , chunks=None    , data=np.full((ny,),0.,np.float64)    )
                hfw.create_dataset( f'Pkz/{scalar}'        , shape=(ny,nkz) , dtype=np.float64 , chunks=(1,nkz) , data=np.full((ny,nkz),0.,np.float64) )
                hfw.create_dataset( f'Pf/{scalar}'         , shape=(ny,nf)  , dtype=np.float64 , chunks=(1,nf)  , data=np.full((ny,nf),0.,np.float64)  )
            
            ## Initialize datasets 1D [y] mean
            for scalar in avg_Re.dtype.names:
                hfw.create_dataset( f'avg/Re/{scalar}', shape=(ny,), dtype=np.float64, chunks=None, data=np.full((ny,),0.,np.float64) )
            for scalar in avg_Fv.dtype.names:
                hfw.create_dataset( f'avg/Fv/{scalar}', shape=(ny,), dtype=np.float64, chunks=None, data=np.full((ny,),0.,np.float64) )
    
    self.comm.Barrier()
    
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## Collectively write covariance,Pkz,Pf
        for scalar in scalars:
            dset = hfw[f'covariance/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = covariance[scalar][:]
            dset = hfw[f'Pkz/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Pkz[scalar][:,:]
            dset = hfw[f'Pf/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Pf[scalar][:,:]
        
        ## Collectively write 1D [y] avgs (Reynolds,Favre)
        for scalar in avg_Re.dtype.names:
            dset = hfw[f'avg/Re/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = avg_Re[scalar][:]
        for scalar in avg_Fv.dtype.names:
            dset = hfw[f'avg/Fv/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = avg_Fv[scalar][:]
    
    ## Report file contents
    self.comm.Barrier()
    if (self.rank==0):
        even_print( os.path.basename(fn_h5_out) , f'{(os.path.getsize(fn_h5_out)/1024**2):0.1f} [MB]' )
        print(72*'-')
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : rgd.calc_turb_cospectrum_xpln() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import sys
import timeit

import numpy as np
from tqdm import tqdm

from .bl import (
    calc_d1,
    calc_d2,
    calc_d3,
    calc_d99_1d,
    calc_dRC,
    calc_I2,
    calc_I3,
)
from .compressible_transform import (
    comp_transform_GFM,
    comp_transform_TL,
    comp_transform_VD,
    comp_transform_VIPL,
)
from .utils import even_print, format_time_string

'''
========================================================================
Compressibility transformations
========================================================================
'''

# ======================================================================

def _calc_VDII(self, **kwargs):
    '''
    Transform [cf,Reθ] to 'incompressible' analog according to Van Driest (1956)
    - Fc (compressibility factor on cf)
    - Reδ2 = (μe/μw)·Reθ
    - cfi = Fc·cf
    Often referred to as 'Van Driest II' compressibility transform
    -----
    - Van Driest (1956) 'The Problem of Aerodynamic Heating'
      https://web.stanford.edu/~jurzay/ME356_files/vandriest_aeroheating.pdf
    - White (2006) 'Viscous Fluid Flow' 7-7 and 7-8 (p.547-556)
    '''
    
    verbose   = kwargs.get('verbose',True)
    adiabatic = kwargs.get('adiabatic',False) ## compute Fc as special case of adiabatic
    
    if verbose: print('\n'+'ztmd.calc_VDII()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if verbose: even_print('adiabatic',str(adiabatic))
    
    T_wall   = np.copy( self['data_1Dx/T_wall'][()]   )
    T_edge   = np.copy( self['data_1Dx/T_edge'][()]   )
    #rho_edge = np.copy( self['data_1Dx/rho_edge'][()] )
    #u_edge   = np.copy( self['data_1Dx/u_edge'][()]   )
    
    cf       = np.copy( self['data_1Dx/cf'][()]       )
    #tau_wall = np.copy( self['data_1Dx/tau_wall'][()] )
    
    if adiabatic:
        Fc = np.copy( ((T_wall / T_edge)-1.) / (np.arcsin( (1.-(T_edge/T_wall))**0.5 ))**2 )
    else:
        Taw = self.Taw
        A  = np.copy( (Taw/T_edge + T_wall/T_edge - 2) / np.sqrt( (Taw/T_edge + T_wall/T_edge)**2 - 4*T_wall/T_edge ) )
        B  = np.copy( (Taw/T_edge - T_wall/T_edge    ) / np.sqrt( (Taw/T_edge + T_wall/T_edge)**2 - 4*T_wall/T_edge ) )
        Fc = np.copy( (Taw/T_edge - 1. ) / ( np.arcsin(A) + np.arcsin(B) )**2 )
    
    cf_inc = np.copy( Fc * cf )
    
    # ===
    
    if ('data_1Dx/cf_inc' in self): del self['data_1Dx/cf_inc']
    self.create_dataset('data_1Dx/cf_inc', data=cf_inc, chunks=None)
    if verbose: even_print('data_1Dx/cf_inc','%s'%str(cf_inc.shape))
    
    if ('data_1Dx/Fc' in self): del self['data_1Dx/Fc']
    self.create_dataset('data_1Dx/Fc', data=Fc, chunks=None)
    if verbose: even_print('data_1Dx/Fc','%s'%str(Fc.shape))
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_VDII() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    
    return

def _calc_comp_trafo(self, schemes=['VD','VIPL','TL','GFM'], **kwargs):
    '''
    Calculate compressible transformations of data (ZTMD wrapper)
    -----
    'VD'   : Van Driest (1951) ........... https://doi.org/10.2514/8.1895
    'VIPL' : Volpiani et al. (2020) ...... https://doi.org/10.1103/PhysRevFluids.5.052602
    'TL'   : Trettel & Larsson (2016) .... https://doi.org/10.1063/1.4942022
    'GFM'  : Griffin, Fu & Moin (2021) ... https://doi.org/10.1073/pnas.2111144118
    '''
    verbose = kwargs.get('verbose',True)
    
    if verbose: print('\n'+'ztmd.calc_comp_trafo()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if 'VD' in schemes:
        u_VD   = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        
        d1_VD  = np.full((self.nx,),np.nan,dtype=np.float64)
        d2_VD  = np.full((self.nx,),np.nan,dtype=np.float64)
        d3_VD  = np.full((self.nx,),np.nan,dtype=np.float64)
        dRC_VD = np.full((self.nx,),np.nan,dtype=np.float64)
        d99_VD = np.full((self.nx,),np.nan,dtype=np.float64)
        I2_VD  = np.full((self.nx,),np.nan,dtype=np.float64)
        I3_VD  = np.full((self.nx,),np.nan,dtype=np.float64)
        H12_VD = np.full((self.nx,),np.nan,dtype=np.float64)
        H32_VD = np.full((self.nx,),np.nan,dtype=np.float64)
        u_edge_VD = np.full((self.nx,),np.nan,dtype=np.float64)
        u_99_VD   = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'VIPL' in schemes:
        u_VIPL   = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        y_VIPL   = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        
        d1_VIPL  = np.full((self.nx,),np.nan,dtype=np.float64)
        d2_VIPL  = np.full((self.nx,),np.nan,dtype=np.float64)
        d3_VIPL  = np.full((self.nx,),np.nan,dtype=np.float64)
        dRC_VIPL = np.full((self.nx,),np.nan,dtype=np.float64)
        d99_VIPL = np.full((self.nx,),np.nan,dtype=np.float64)
        I2_VIPL  = np.full((self.nx,),np.nan,dtype=np.float64)
        I3_VIPL  = np.full((self.nx,),np.nan,dtype=np.float64)
        H12_VIPL = np.full((self.nx,),np.nan,dtype=np.float64)
        H32_VIPL = np.full((self.nx,),np.nan,dtype=np.float64)
        u_edge_VIPL = np.full((self.nx,),np.nan,dtype=np.float64)
        u_99_VIPL   = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'TL' in schemes:
        u_TL      = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        u_plus_TL = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        y_TL      = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        y_plus_TL = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        
        d1_TL  = np.full((self.nx,),np.nan,dtype=np.float64)
        d2_TL  = np.full((self.nx,),np.nan,dtype=np.float64)
        d3_TL  = np.full((self.nx,),np.nan,dtype=np.float64)
        dRC_TL = np.full((self.nx,),np.nan,dtype=np.float64)
        d99_TL = np.full((self.nx,),np.nan,dtype=np.float64)
        I2_TL  = np.full((self.nx,),np.nan,dtype=np.float64)
        I3_TL  = np.full((self.nx,),np.nan,dtype=np.float64)
        H12_TL = np.full((self.nx,),np.nan,dtype=np.float64)
        H32_TL = np.full((self.nx,),np.nan,dtype=np.float64)
        u_edge_TL = np.full((self.nx,),np.nan,dtype=np.float64)
        u_99_TL   = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'GFM' in schemes:
        u_GFM      = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        u_plus_GFM = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        y_GFM      = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        y_plus_GFM = np.full((self.nx,self.ny),np.nan,dtype=np.float64)
        
        d1_GFM  = np.full((self.nx,),np.nan,dtype=np.float64)
        d2_GFM  = np.full((self.nx,),np.nan,dtype=np.float64)
        d3_GFM  = np.full((self.nx,),np.nan,dtype=np.float64)
        dRC_GFM = np.full((self.nx,),np.nan,dtype=np.float64)
        d99_GFM = np.full((self.nx,),np.nan,dtype=np.float64)
        I2_GFM  = np.full((self.nx,),np.nan,dtype=np.float64)
        I3_GFM  = np.full((self.nx,),np.nan,dtype=np.float64)
        H12_GFM = np.full((self.nx,),np.nan,dtype=np.float64)
        H32_GFM = np.full((self.nx,),np.nan,dtype=np.float64)
        u_edge_GFM = np.full((self.nx,),np.nan,dtype=np.float64)
        u_99_GFM   = np.full((self.nx,),np.nan,dtype=np.float64)
    
    ## Read data
    u    = np.copy( self['data/u'][()].T    )
    u_Fv = np.copy( self['data/u_Fv'][()].T )
    rho  = np.copy( self['data/rho'][()].T  )
    mu   = np.copy( self['data/mu'][()].T   )
    
    u_tau  = np.copy( self['data_1Dx/u_tau'][()]  )
    j_edge = np.copy( self['data_1Dx/j_edge'][()] )
    
    ddy_rho   = np.copy( self['data/ddy_rho'][()].T   )
    ddy_mu    = np.copy( self['data/ddy_mu'][()].T    )
    ddy_u_Fv  = np.copy( self['data/ddy_u_Fv'][()].T  )
    r_uII_vII = np.copy( self['data/r_uII_vII'][()].T )
    
    if verbose:
        progress_bar = tqdm(
            total=self.nx,
            ncols=100,
            desc='calc_comp_trafo()',
            leave=False,
            file=sys.stdout,
            )
    
    ## Main loop
    for i in range(self.nx):
        
        je = j_edge[i]
        
        if 'VD' in schemes:
            u_VD_ = comp_transform_VD(u[i,:], rho[i,:])
            u_VD[i,:] = u_VD_
            y_VD_     = self.y
            
            d1_VD[i]  = calc_d1(y_VD_, u_VD_, j_edge=je)
            d2_VD[i]  = calc_d2(y_VD_, u_VD_, j_edge=je)
            d3_VD[i]  = calc_d3(y_VD_, u_VD_, j_edge=je)
            dRC_VD[i] = calc_dRC(y_VD_, u_VD_, u_tau=u_tau[i], j_edge=je)
            I2_VD[i]  = calc_I2(y_VD_, u_VD_, u_tau=u_tau[i], j_edge=je)
            I3_VD[i]  = calc_I3(y_VD_, u_VD_, u_tau=u_tau[i], j_edge=je)
            H12_VD[i] = d1_VD[i] / d2_VD[i]
            H32_VD[i] = d3_VD[i] / d2_VD[i]
            
            d99_VD[i] = calc_d99_1d(y=y_VD_, u=u_VD_, j_edge=je, interp_kind='cubic')
            u_edge_VD[i] = u_VD[i,je]
            u_99_VD[i] = 0.99*u_VD[i,je]
        
        if 'VIPL' in schemes:
            u_VIPL_, y_VIPL_ = comp_transform_VIPL(self.y, u[i,:], rho[i,:], mu[i,:])
            u_VIPL[i,:] = u_VIPL_
            y_VIPL[i,:] = y_VIPL_
            
            d1_VIPL[i]  = calc_d1(y_VIPL_, u_VIPL_, j_edge=je)
            d2_VIPL[i]  = calc_d2(y_VIPL_, u_VIPL_, j_edge=je)
            d3_VIPL[i]  = calc_d3(y_VIPL_, u_VIPL_, j_edge=je)
            dRC_VIPL[i] = calc_dRC(y_VIPL_, u_VIPL_, u_tau=u_tau[i], j_edge=je)
            I2_VIPL[i]  = calc_I2(y_VIPL_, u_VIPL_, u_tau=u_tau[i], j_edge=je)
            I3_VIPL[i]  = calc_I3(y_VIPL_, u_VIPL_, u_tau=u_tau[i], j_edge=je)
            H12_VIPL[i] = d1_VIPL[i] / d2_VIPL[i]
            H32_VIPL[i] = d3_VIPL[i] / d2_VIPL[i]
            
            d99_VIPL[i] = calc_d99_1d(y=y_VIPL_, u=u_VIPL_, j_edge=je, interp_kind='cubic')
            u_edge_VIPL[i] = u_VIPL_[je]
            u_99_VIPL[i] = 0.99*u_VIPL[i,je]
        
        if 'TL' in schemes: ## Trettel & Larsson
            u_TL_, y_TL_, u_plus_TL_, y_plus_TL_ = comp_transform_TL(
                                                        y=self.y,
                                                        u=u[i,:],
                                                        utau=u_tau[i],
                                                        rho=rho[i,:],
                                                        mu=mu[i,:],
                                                        ddy_mu=ddy_mu[i,:],
                                                        ddy_rho=ddy_rho[i,:],
                                                        )
            
            u_TL[i,:]      = u_TL_
            u_plus_TL[i,:] = u_plus_TL_
            y_TL[i,:]      = y_TL_
            y_plus_TL[i,:] = y_plus_TL_
            
            d1_TL[i]  = calc_d1(y_TL_, u_TL_, j_edge=je)
            d2_TL[i]  = calc_d2(y_TL_, u_TL_, j_edge=je)
            d3_TL[i]  = calc_d3(y_TL_, u_TL_, j_edge=je)
            dRC_TL[i] = calc_dRC(y_TL_, u_TL_, u_tau=u_tau[i], j_edge=je)
            I2_TL[i]  = calc_I2(y_TL_, u_TL_, u_tau=u_tau[i], j_edge=je)
            I3_TL[i]  = calc_I3(y_TL_, u_TL_, u_tau=u_tau[i], j_edge=je)
            H12_TL[i] = d1_TL[i] / d2_TL[i]
            H32_TL[i] = d3_TL[i] / d2_TL[i]
            
            d99_TL[i] = calc_d99_1d(y=y_TL_, u=u_TL_, j_edge=je, interp_kind='cubic')
            u_edge_TL[i] = u_TL_[je]
            u_99_TL[i] = 0.99*u_TL[i,je]
        
        if 'GFM' in schemes: ## Griffin, Fu & Moin
            u_GFM_, y_GFM_, u_plus_GFM_, y_plus_GFM_ = comp_transform_GFM(
                                                            y=self.y,
                                                            uFv=u_Fv[i,:],
                                                            rho=rho[i,:],
                                                            mu=mu[i,:],
                                                            r_uII_vII=r_uII_vII[i,:],
                                                            utau=u_tau[i],
                                                            ddy_u_Fv=ddy_u_Fv[i,:],
                                                            )
            
            u_GFM[i,:]      = u_GFM_
            u_plus_GFM[i,:] = u_plus_GFM_
            y_GFM[i,:]      = y_GFM_
            y_plus_GFM[i,:] = y_plus_GFM_
            
            d1_GFM[i]  = calc_d1(y_GFM_, u_GFM_, j_edge=je)
            d2_GFM[i]  = calc_d2(y_GFM_, u_GFM_, j_edge=je)
            d3_GFM[i]  = calc_d3(y_GFM_, u_GFM_, j_edge=je)
            dRC_GFM[i] = calc_dRC(y_GFM_, u_GFM_, u_tau=u_tau[i], j_edge=je)
            I2_GFM[i]  = calc_I2(y_GFM_, u_GFM_, u_tau=u_tau[i], j_edge=je)
            I3_GFM[i]  = calc_I3(y_GFM_, u_GFM_, u_tau=u_tau[i], j_edge=je)
            H12_GFM[i] = d1_GFM[i] / d2_GFM[i]
            H32_GFM[i] = d3_GFM[i] / d2_GFM[i]
            
            d99_GFM[i] = calc_d99_1d(y=y_GFM_, u=u_GFM_, j_edge=je, interp_kind='cubic')
            u_edge_GFM[i] = u_GFM_[je]
            u_99_GFM[i] = 0.99*u_GFM[i,je]
        
        if verbose: progress_bar.update()
    if verbose: progress_bar.close()
    
    # ===
    
    ## Helper func for writing 1D datasets to data_1Dx/
    def _write_1D(name, arr):
        if name in self: ## Delete if exists
            del self[name]
        self.create_dataset(f'data_1Dx/{name}', data=arr, chunks=None) ## Write
        if verbose:
            even_print(f'data_1Dx/{name}', str(arr.shape))
    
    ## Helper func for writing 2D datasets to data/
    def _write_2D(name, arr):
        if name in self: ## Delete if exists
            del self[name]
        self.create_dataset(f'data/{name}', data=arr.T, chunks=None) ## Write
        if verbose:
            even_print(f'data/{name}', str(arr.shape))
    
    if 'VD' in schemes: ## van-Driest
        _write_2D('u_VD',u_VD)
        
        _write_1D('u_edge_VD' , u_edge_VD )
        _write_1D('u_99_VD'   , u_99_VD   )
        
        _write_1D('d1_VD'  , d1_VD  )
        _write_1D('d2_VD'  , d2_VD  )
        _write_1D('d3_VD'  , d3_VD  )
        _write_1D('dRC_VD' , dRC_VD )
        _write_1D('d99_VD' , d99_VD )
        _write_1D('I2_VD'  , I2_VD  )
        _write_1D('I3_VD'  , I3_VD  )
        _write_1D('H12_VD' , H12_VD )
        _write_1D('H32_VD' , H32_VD )
    
    if 'VIPL' in schemes: ## Volpiani et al.
        _write_2D('u_VIPL',u_VIPL)
        _write_2D('y_VIPL',y_VIPL)
        
        _write_1D('u_edge_VIPL' , u_edge_VIPL )
        _write_1D('u_99_VIPL'   , u_99_VIPL   )
        
        _write_1D('d1_VIPL'  , d1_VIPL  )
        _write_1D('d2_VIPL'  , d2_VIPL  )
        _write_1D('d3_VIPL'  , d3_VIPL  )
        _write_1D('dRC_VIPL' , dRC_VIPL )
        _write_1D('d99_VIPL' , d99_VIPL )
        _write_1D('I2_VIPL'  , I2_VIPL  )
        _write_1D('I3_VIPL'  , I3_VIPL  )
        _write_1D('H12_VIPL' , H12_VIPL )
        _write_1D('H32_VIPL' , H32_VIPL )
    
    if 'TL' in schemes: ## Trettel & Larsson
        _write_2D('u_TL'      , u_TL      )
        _write_2D('u_plus_TL' , u_plus_TL )
        _write_2D('y_TL'      , y_TL      )
        _write_2D('y_plus_TL' , y_plus_TL )
        
        _write_1D('u_edge_TL' , u_edge_TL )
        _write_1D('u_99_TL'   , u_99_TL   )
        
        _write_1D('d1_TL'  , d1_TL  )
        _write_1D('d2_TL'  , d2_TL  )
        _write_1D('d3_TL'  , d3_TL  )
        _write_1D('dRC_TL' , dRC_TL )
        _write_1D('d99_TL' , d99_TL )
        _write_1D('I2_TL'  , I2_TL  )
        _write_1D('I3_TL'  , I3_TL  )
        _write_1D('H12_TL' , H12_TL )
        _write_1D('H32_TL' , H32_TL )
    
    if 'GFM' in schemes: ## Griffin,Fu & Moin
        _write_2D('u_GFM'      , u_GFM      )
        _write_2D('u_plus_GFM' , u_plus_GFM )
        _write_2D('y_GFM'      , y_GFM      )
        _write_2D('y_plus_GFM' , y_plus_GFM )
        
        _write_1D('u_edge_GFM' , u_edge_GFM )
        _write_1D('u_99_GFM'   , u_99_GFM   )
        
        _write_1D('d1_GFM'  , d1_GFM  )
        _write_1D('d2_GFM'  , d2_GFM  )
        _write_1D('d3_GFM'  , d3_GFM  )
        _write_1D('dRC_GFM' , dRC_GFM )
        _write_1D('d99_GFM' , d99_GFM )
        _write_1D('I2_GFM'  , I2_GFM  )
        _write_1D('I3_GFM'  , I3_GFM  )
        _write_1D('H12_GFM' , H12_GFM )
        _write_1D('H32_GFM' , H32_GFM )
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_comp_trafo() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import matplotlib as mpl

'''
========================================================================
matplotlib environment initializer
========================================================================
'''

# ======================================================================

def set_mpl_env(**kwargs):
    '''
    Set matplotlib global presets
    '''
    fontsize = kwargs.get('fontsize',9)
    usetex   = kwargs.get('usetex',True)
    sfac     = kwargs.get('sfac',1) ## global scale factor
    
    fontsize *= sfac ## scale font size by global scale factor
    axesAndTickWidth = 0.5*sfac
    
    mpl.rcParams['text.usetex'] = usetex
    
    if usetex:
        preamble_opts = [
                        r'\usepackage{amsmath}',
                        r'\usepackage{amssymb}',
                        r'\usepackage{gensymb}',
                        #r'\usepackage{graphicx}', ## not supported
                        r'\usepackage{newtxtext}', ## Times
                        r'\usepackage{newtxmath}', ## Times Math
                        r'\usepackage{xfrac}',
                        ]
        mpl.rcParams['text.latex.preamble'] = '\n'.join(preamble_opts)
        
        mpl.rcParams['font.family'] = 'serif'
        #mpl.rcParams['font.serif']  = 'Computer Modern Roman'
        mpl.rcParams['font.serif']  = 'Times'
    
    else:
        pass
    
    mpl.rcParams['xtick.major.size']  = 2.5*sfac
    mpl.rcParams['xtick.major.width'] = axesAndTickWidth
    mpl.rcParams['xtick.minor.size']  = 1.5*sfac
    mpl.rcParams['xtick.minor.width'] = axesAndTickWidth
    mpl.rcParams['xtick.direction']   = 'in'
    
    mpl.rcParams['ytick.major.size']  = 2.5*sfac
    mpl.rcParams['ytick.major.width'] = axesAndTickWidth
    mpl.rcParams['ytick.minor.size']  = 1.5*sfac
    mpl.rcParams['ytick.minor.width'] = axesAndTickWidth
    mpl.rcParams['ytick.direction']   = 'in'
    
    mpl.rcParams['xtick.labelsize'] = fontsize
    mpl.rcParams['ytick.labelsize'] = fontsize
    
    mpl.rcParams['xtick.major.pad'] = 3.0*sfac
    mpl.rcParams['xtick.minor.pad'] = 3.0*sfac
    mpl.rcParams['ytick.major.pad'] = 3.0*sfac
    mpl.rcParams['ytick.minor.pad'] = 3.0*sfac
    
    mpl.rcParams['lines.linewidth']       = 1.0
    mpl.rcParams['lines.linestyle']       = 'solid'
    mpl.rcParams['lines.marker']          = 'None' #'o'
    mpl.rcParams['lines.markersize']      = 1.2
    mpl.rcParams['lines.markeredgewidth'] = 0.
    
    mpl.rcParams['axes.linewidth'] = axesAndTickWidth
    mpl.rcParams['axes.labelpad']  = 2.0
    mpl.rcParams['axes.titlesize'] = fontsize
    mpl.rcParams['axes.labelsize'] = fontsize
    mpl.rcParams['axes.formatter.use_mathtext'] = True
    
    mpl.rcParams['legend.fontsize']      = fontsize
    mpl.rcParams['legend.shadow']        = False
    mpl.rcParams['legend.borderpad']     = 0.3
    mpl.rcParams['legend.framealpha']    = 1.0
    mpl.rcParams['legend.edgecolor']     = 'inherit'
    mpl.rcParams['legend.handlelength']  = 1.5
    mpl.rcParams['legend.handletextpad'] = 0.3
    mpl.rcParams['legend.borderaxespad'] = 0.7
    mpl.rcParams['legend.columnspacing'] = 0.5
    mpl.rcParams['legend.fancybox']      = False
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pytest

from turbx import gradient


@pytest.mark.parametrize( "acc, edge, tol", [
    (2, "half", 1e-2),
    (4, "half", 1e-3),
    (6, "half", 1e-6),
    (8, "half", 1e-7),
    (2, "full", 1e-2),
    (4, "full", 1e-3),
    (6, "full", 1e-6),
    (8, "full", 1e-7),
    ],
)
def test_deriv_1d_o1(acc, edge, tol):
    '''
    Test first derivative against analytic sin -> cos.
    '''
    nx = 500
    x = np.linspace(0, 4*np.pi, nx, dtype=np.float64)
    u = np.sin(x)
    exact = np.cos(x)
    num = gradient(
        u=u,
        x=x,
        d=1,
        axis=0,
        acc=acc,
        edge_stencil=edge,
    )
    np.testing.assert_allclose(exact, num, rtol=tol, atol=tol)



import timeit

import numpy as np
from tqdm import tqdm

from .gradient import gradient
from .utils import even_print

# ======================================================================

def get_metric_tensor_2d(x2d, y2d, acc=2, edge_stencil='full', **kwargs):
    '''
    Compute the grid metric tensor (inverse of grid Jacobian) for a 2D grid
    -----
    Computational Fluid Mechanics and Heat Transfer (2012) Pletcher, Tannehill, Anderson
    p.266-270, 335-337, 652
    '''
    
    verbose = kwargs.get('verbose',False)
    no_warn = kwargs.get('no_warn',False)
    
    if not isinstance(x2d, np.ndarray):
        raise ValueError('x2d should be of type np.ndarray')
    if not isinstance(y2d, np.ndarray):
        raise ValueError('y2d should be of type np.ndarray')
    
    if (x2d.ndim!=2):
        raise ValueError('x2d should have ndim=2 (xy)')
    if (y2d.ndim!=2):
        raise ValueError('y2d should have ndim=2 (xy)')
    
    if not (x2d.shape==y2d.shape):
        raise ValueError('x2d.shape!=y2d.shape')
    
    nx,ny = x2d.shape
    
    ## the 'computational' grid (unit Cartesian)
    ## --> [x_comp,y_comp]= [ξ,η] = [q1,q2]
    #x_comp = np.arange(nx, dtype=np.float64)
    #y_comp = np.arange(ny, dtype=np.float64)
    x_comp = 1.
    y_comp = 1.
    
    # === get Jacobian :: ∂(x,y)/∂(q1,q2)
    
    t_start = timeit.default_timer()
    
    dxdx = gradient(x2d, x_comp, axis=0, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dydx = gradient(y2d, x_comp, axis=0, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dxdy = gradient(x2d, y_comp, axis=1, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dydy = gradient(y2d, y_comp, axis=1, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    
    J = np.stack((np.stack((dxdx, dydx), axis=2),
                  np.stack((dxdy, dydy), axis=2)), axis=3)
    
    t_delta = timeit.default_timer() - t_start
    if verbose: tqdm.write( even_print('get J','%0.3f [s]'%(t_delta,), s=True) )
    
    # === get metric tensor M = J^-1 = ∂(q1,q2)/∂(x,y) = ∂(ξ,η)/∂(x,y)
    
    if False: ## method 1
        
        t_start = timeit.default_timer()
        
        M = np.linalg.inv(J)
        
        # M_bak = np.copy(M)
        # M = np.zeros((nx,ny,2,2),dtype=np.float64)
        # for i in range(nx):
        #     for j in range(ny):
        #         M[i,j,:,:] = sp.linalg.inv( J[i,j,:,:] )
        # np.testing.assert_allclose(M_bak, M, atol=1e-12, rtol=1e-12)
        # print('check passed')
        
        t_delta = timeit.default_timer() - t_start
        if verbose: tqdm.write( even_print('get M','%0.3f [s]'%(t_delta,), s=True) )
    
    if True: ## method 2
        
        if ('M' in locals()):
            M_bak = np.copy(M)
            M = None; del M
        
        t_start = timeit.default_timer()
        
        ## Jacobian determinant
        Jac_det = dxdx*dydy - dydx*dxdy
        
        # Jac_det_bak = np.copy(Jac_det)
        # Jac_det = None; del Jac_det
        # Jac_det = np.linalg.det(J)
        # np.testing.assert_allclose(Jac_det, Jac_det_bak, atol=1e-14, rtol=1e-14)
        # print('check passed')
        
        M = np.zeros((nx,ny,2,2), dtype=np.float64)
        M[:,:,0,0] = +dydy / Jac_det ## ξ_x
        M[:,:,0,1] = -dxdy / Jac_det ## ξ_y
        M[:,:,1,0] = -dydx / Jac_det ## η_x
        M[:,:,1,1] = +dxdx / Jac_det ## η_y
        
        t_delta = timeit.default_timer() - t_start
        if verbose: tqdm.write( even_print('get M','%0.3f [s]'%(t_delta,), s=True) )
        
        if ('M_bak' in locals()):
            np.testing.assert_allclose(M[:,:,0,0], M_bak[:,:,0,0], atol=1e-14, rtol=1e-14)
            print('check passed: ξ_x')
            np.testing.assert_allclose(M[:,:,0,1], M_bak[:,:,0,1], atol=1e-14, rtol=1e-14)
            print('check passed: ξ_y')
            np.testing.assert_allclose(M[:,:,1,0], M_bak[:,:,1,0], atol=1e-14, rtol=1e-14)
            print('check passed: η_x')
            np.testing.assert_allclose(M[:,:,1,1], M_bak[:,:,1,1], atol=1e-14, rtol=1e-14)
            print('check passed: η_y')
            np.testing.assert_allclose(M, M_bak, atol=1e-14, rtol=1e-14)
            print('check passed: M')
    
    return M

def get_metric_tensor_3d(x3d, y3d, z3d, acc=2, edge_stencil='full', **kwargs):
    '''
    Compute the grid metric tensor (inverse of grid Jacobian) for a 3D grid
    -----
    Computational Fluid Mechanics and Heat Transfer (2012) Pletcher, Tannehill, Anderson
    p.266-270, 335-337, 652
    '''
    
    verbose = kwargs.get('verbose',False)
    no_warn = kwargs.get('no_warn',False)
    
    if not isinstance(x3d, np.ndarray):
        raise ValueError('x3d should be of type np.ndarray')
    if not isinstance(y3d, np.ndarray):
        raise ValueError('y3d should be of type np.ndarray')
    if not isinstance(z3d, np.ndarray):
        raise ValueError('z3d should be of type np.ndarray')
    
    if (x3d.ndim!=3):
        raise ValueError('x3d should have ndim=3 (xyz)')
    if (y3d.ndim!=3):
        raise ValueError('y3d should have ndim=3 (xyz)')
    if (z3d.ndim!=3):
        raise ValueError('z3d should have ndim=3 (xyz)')
    
    if not (x3d.shape==y3d.shape):
        raise ValueError('x3d.shape!=y3d.shape')
    if not (y3d.shape==z3d.shape):
        raise ValueError('y3d.shape!=z3d.shape')
    
    nx,ny,nz = x3d.shape
    
    ## the 'computational' grid (unit Cartesian)
    ## --> [x_comp,y_comp,z_comp ]= [ξ,η,ζ] = [q1,q2,q3]
    #x_comp = np.arange(nx, dtype=np.float64)
    #y_comp = np.arange(ny, dtype=np.float64)
    #z_comp = np.arange(nz, dtype=np.float64)
    x_comp = 1.
    y_comp = 1.
    z_comp = 1.
    
    # === get Jacobian :: ∂(x,y,z)/∂(q1,q2,q3)
    
    t_start = timeit.default_timer()
    
    dxdx = gradient(x3d, x_comp, axis=0, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dydx = gradient(y3d, x_comp, axis=0, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dzdx = gradient(z3d, x_comp, axis=0, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    
    dxdy = gradient(x3d, y_comp, axis=1, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dydy = gradient(y3d, y_comp, axis=1, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dzdy = gradient(z3d, y_comp, axis=1, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    
    dxdz = gradient(x3d, z_comp, axis=2, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dydz = gradient(y3d, z_comp, axis=2, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    dzdz = gradient(z3d, z_comp, axis=2, d=1, acc=acc, edge_stencil=edge_stencil, no_warn=no_warn)
    
    J = np.stack((np.stack((dxdx, dydx, dzdx), axis=3),
                  np.stack((dxdy, dydy, dzdy), axis=3),
                  np.stack((dxdz, dydz, dzdz), axis=3)), axis=4)
    
    t_delta = timeit.default_timer() - t_start
    if verbose: tqdm.write( even_print('get J','%0.3f [s]'%(t_delta,), s=True) )
    
    # === get metric tensor M = J^-1 = ∂(q1,q2,q3)/∂(x,y,z) = ∂(ξ,η,ζ)/∂(x,y,z)
    
    if False: ## method 1
        
        t_start = timeit.default_timer()
        
        M = np.linalg.inv(J)
        
        # M_bak = np.copy(M)
        # for i in range(nx):
        #     for j in range(ny):
        #         for k in range(nz):
        #             M[i,j,k,:,:] = sp.linalg.inv( J[i,j,k,:,:] )
        # np.testing.assert_allclose(M_bak, M, atol=1e-12, rtol=1e-12)
        # print('check passed')
        
        t_delta = timeit.default_timer() - t_start
        if verbose: tqdm.write( even_print('get M','%0.3f [s]'%(t_delta,), s=True) )
    
    if True: ## method 2
        
        if ('M' in locals()):
            M_bak = np.copy(M)
            M = None; del M
        
        t_start = timeit.default_timer()
        
        a = J[:,:,:,0,0]
        b = J[:,:,:,0,1]
        c = J[:,:,:,0,2]
        d = J[:,:,:,1,0]
        e = J[:,:,:,1,1]
        f = J[:,:,:,1,2]
        g = J[:,:,:,2,0]
        h = J[:,:,:,2,1]
        i = J[:,:,:,2,2]
        
        # a = J[:,:,:,0,0]
        # b = J[:,:,:,1,0]
        # c = J[:,:,:,2,0]
        # d = J[:,:,:,0,1]
        # e = J[:,:,:,1,1]
        # f = J[:,:,:,2,1]
        # g = J[:,:,:,0,2]
        # h = J[:,:,:,1,2]
        # i = J[:,:,:,2,2]
        
        Jac_det = ( + a*e*i
                    + b*f*g
                    + c*d*h
                    - c*e*g
                    - b*d*i
                    - a*f*h )
        
        M = np.zeros((nx,ny,nz,3,3), dtype=np.float64)
        M[:,:,:,0,0] = +( dydy * dzdz - dydz * dzdy ) / Jac_det ## ξ_x
        M[:,:,:,0,1] = -( dxdy * dzdz - dxdz * dzdy ) / Jac_det ## ξ_y
        M[:,:,:,0,2] = +( dxdy * dydz - dxdz * dydy ) / Jac_det ## ξ_z
        M[:,:,:,1,0] = -( dydx * dzdz - dydz * dzdx ) / Jac_det ## η_x
        M[:,:,:,1,1] = +( dxdx * dzdz - dxdz * dzdx ) / Jac_det ## η_y
        M[:,:,:,1,2] = -( dxdx * dydz - dxdz * dydx ) / Jac_det ## η_z
        M[:,:,:,2,0] = +( dydx * dzdy - dydy * dzdx ) / Jac_det ## ζ_x
        M[:,:,:,2,1] = -( dxdx * dzdy - dxdy * dzdx ) / Jac_det ## ζ_y
        M[:,:,:,2,2] = +( dxdx * dydy - dxdy * dydx ) / Jac_det ## ζ_z
        
        t_delta = timeit.default_timer() - t_start
        if verbose: tqdm.write( even_print('get M','%0.3f [s]'%(t_delta,), s=True) )
        
        if ('M_bak' in locals()):
            np.testing.assert_allclose(M[:,:,:,0,0], M_bak[:,:,:,0,0], atol=1e-14, rtol=1e-14)
            print('check passed: ξ_x')
            np.testing.assert_allclose(M[:,:,:,0,1], M_bak[:,:,:,0,1], atol=1e-14, rtol=1e-14)
            print('check passed: ξ_y')
            np.testing.assert_allclose(M[:,:,:,0,2], M_bak[:,:,:,0,2], atol=1e-14, rtol=1e-14)
            print('check passed: ξ_z')
            np.testing.assert_allclose(M[:,:,:,1,0], M_bak[:,:,:,1,0], atol=1e-14, rtol=1e-14)
            print('check passed: η_x')
            np.testing.assert_allclose(M[:,:,:,1,1], M_bak[:,:,:,1,1], atol=1e-14, rtol=1e-14)
            print('check passed: η_y')
            np.testing.assert_allclose(M[:,:,:,1,2], M_bak[:,:,:,1,2], atol=1e-14, rtol=1e-14)
            print('check passed: η_z')
            np.testing.assert_allclose(M[:,:,:,2,0], M_bak[:,:,:,2,0], atol=1e-14, rtol=1e-14)
            print('check passed: ζ_x')
            np.testing.assert_allclose(M[:,:,:,2,1], M_bak[:,:,:,2,1], atol=1e-14, rtol=1e-14)
            print('check passed: ζ_y')
            np.testing.assert_allclose(M[:,:,:,2,2], M_bak[:,:,:,2,2], atol=1e-14, rtol=1e-14)
            print('check passed: ζ_z')
            np.testing.assert_allclose(M, M_bak, atol=1e-14, rtol=1e-14)
            print('check passed: M')
    
    return M



import os
import sys
import timeit
import traceback
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
from tqdm import tqdm

from .confidence_interval import calc_var_bmbc, confidence_interval_unbiased
from .h5 import h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_mean_uncertainty_BMBC(self, **kwargs):
    '''
    calculate the uncertainty of the mean using the
    "Batch Means and Batch Correlations" (BMBC) method outlined in
    §4 of https://doi.org/10.1016/j.jcp.2017.07.005
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'spd.calc_mean_uncertainty_BMBC()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    h5py_is_mpi_build = h5py.h5.get_config().mpi
    if not h5py_is_mpi_build:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    ri = kwargs.get('ri',1)
    rj = kwargs.get('rj',1)
    rt = kwargs.get('rt',1)
    
    ## #n_threads = kwargs.get('n_threads',1)
    ## try:
    ##     n_threads = int(os.environ.get('OMP_NUM_THREADS'))
    ## except TypeError: ## not set
    ##     n_threads = os.cpu_count()
    
    fn_h5_out = kwargs.get('fn_h5_out',None) ## filename for output HDF5 (.h5) file
    
    confidence = kwargs.get('confidence',0.99)
    M          = kwargs.get('M',100) ## batch size
    
    if not isinstance(M, (int,np.int32,np.int64)):
        raise ValueError("'M' should be an int")
    if (M < 1):
        raise ValueError('M < 1')
    
    if (self.nt%M!=0):
        if verbose: print(f'nt = {self.nt:d}')
        if verbose: print(f'M = {M:d}')
        raise ValueError('nt%M!=0')
    
    if not isinstance(confidence, (float,np.float32,np.float64)):
        raise ValueError("'confidence' should be a float")
    if (confidence <= 0.) or (confidence >= 1.):
        raise ValueError('confidence should be between 0,1')
    
    ## check data distribution
    if (rj!=1):
        raise AssertionError('rj!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    if (ri*rj*rt != self.n_ranks):
        raise AssertionError('ri*rj*rt != self.n_ranks')
    if (ri>self.ni):
        raise AssertionError('ri>self.ni')
    if (self.ni%ri!=0):
        raise AssertionError('ni currently needs to be divisible by the n ranks')
    
    ## distribute data over i/[x]
    ril_ = np.array_split(np.arange(self.ni,dtype=np.int64),min(ri,self.ni))
    ril = [[b[0],b[-1]+1] for b in ril_ ]
    ri1,ri2 = ril[self.rank]
    nir = ri2 - ri1
    
    ## output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        #fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fname_out_h5_base = fname_root+'_uncertainty_bmbc.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fname_out_h5_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'     , self.fname )
    if verbose: even_print( 'fn_h5_out' , fn_h5_out  )
    if verbose: print(72*'-')
    if verbose: even_print( 'ni' , f'{self.ni:d}' )
    if verbose: even_print( 'nj' , f'{self.nj:d}' )
    if verbose: even_print( 'nt' , f'{self.nt:d}' )
    if verbose: print(72*'-')
    if verbose: even_print('n ranks', f'{self.n_ranks:d}' )
    #if verbose: even_print('n threads', f'{n_threads:d}' )
    if verbose: print(72*'-')
    if verbose: even_print('M', f'{M:d}' )
    if verbose: even_print('confidence level', f'{confidence:0.4f}' )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    # ===
    
    ## the data dictionary to be pickled or written to .h5 later
    data = {}
    
    ## freestream data
    data['lchar']   = self.lchar
    data['U_inf']   = self.U_inf
    data['rho_inf'] = self.rho_inf
    data['T_inf']   = self.T_inf
    data['mu_inf']  = self.mu_inf
    data['p_inf']   = self.p_inf
    data['Ma']      = self.Ma
    data['Pr']      = self.Pr
    
    ## read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    data['x'] = x
    data['z'] = z
    data['t'] = t
    
    nx = self.ni
    ni = self.ni
    nz = self.nj
    nj = self.nj
    data['ni'] = ni
    data['nx'] = nx
    data['nj'] = nj
    data['nz'] = nz
    
    nt = self.nt
    data['nt'] = nt
    
    ## assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    
    ## get Δt, dimensional [s]
    if hasattr(self,'dt'):
        dt = self.dt * self.tchar
        np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    else:
        dt = t[1] - t[0] ## already dimensional
    
    if hasattr(self,'duration'):
        t_meas = self.duration * self.tchar
        np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    else:
        t_meas = t[-1] - t[0] ## already dimensional
        self.duration = t_meas / self.tchar ## non-dimensionalize for attribute
    
    zrange = z.max() - z.min()
    
    data['t'] = t
    data['dt'] = dt
    data['dz'] = dz0
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        print(72*'-')
    
    # ==============================================================
    # prepare buffers, maps, etc.
    # ==============================================================
    
    # n_lags = 2*nt-1
    
    data['M'] = M
    data['confidence'] = confidence
    
    ## HARDCODE
    scalars = [ 'tau_uy', 'u_tau', ]
    #scalars = [ 'tau_uy', ]
    
    K_full = nt // M
    
    if (K_full<10):
        raise ValueError('nt//M<10')
    
    #Ks = np.arange(3,K_full+1,dtype=np.int64)
    Ks = np.arange(10,K_full+1,dtype=np.int64)
    #Ks = np.copy(Ks[-1:]) ## debug, take last N (full time series)
    Ns = M * Ks
    #if verbose: print(Ns)
    nN = Ns.shape[0]
    
    ## report
    if verbose:
        even_print('n time durations (N)' , f'{nN:d}' )
        #print(72*'-')
    
    data['Ks'] = Ks
    data['Ns'] = Ns
    
    ## rank-local buffers
    data_mean  = np.zeros(shape=(nir,nN)   , dtype={'names':scalars , 'formats':[ np.float64 for sss in scalars ]})
    data_ci    = np.zeros(shape=(nir,nN,2) , dtype={'names':scalars , 'formats':[ np.float64 for sss in scalars ]})
    data_Nsig2 = np.zeros(shape=(nir,nN,)  , dtype={'names':scalars , 'formats':[ np.float64 for sss in scalars ]})
    #data_acor = np.zeros(shape=(nir,n_lags) , dtype={'names':scalars , 'formats':[ np.float64 for sss in scalars ]})
    
    # ==============================================================
    # main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            total=len(scalars)*nir,
            ncols=100,
            desc='BMBC',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    for scalar in scalars:
        
        dset = self[f'data/{scalar}']
        if verbose: tqdm.write(72*'-')
        
        for ii in range(ri1,ri2):
            
            iii = ii - ri1
            
            ## COLLECTIVE read
            self.comm.Barrier()
            t_start = timeit.default_timer()
            if self.usingmpi:
                with dset.collective:
                    dd = np.copy( dset[ii,:,:] )
            else:
                dd = np.copy( dset[ii,:,:] )
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = self.n_ranks * 1 * self.nj * self.nt * dset.dtype.itemsize / 1024**3
            if verbose:
                tqdm.write(even_print(f'read: {scalar}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## assert shape [nz,nt]
            if ( dd.shape != (nj,nt) ):
                raise ValueError
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## cast to double
            dd = np.copy( dd.astype(np.float64) )
            
            ## re-dimensionalize
            if scalar in ['tau_uy','tau_vy','tau_wy',]:
                dd *= self.rho_inf * self.U_inf**2
            elif scalar in ['u_tau','v_tau','w_tau',]:
                dd *= self.U_inf
            elif scalar in ['T',]:
                dd *= self.T_inf
            elif scalar in ['rho',]:
                dd *= self.rho_inf
            elif scalar in ['p',]:
                dd *= self.rho_inf * self.U_inf**2
            else:
                raise ValueError(f"condition needed for redimensionalizing '{scalar}'")
            
            # ======================================================
            
            for iN,N in enumerate(Ns):
                
                ddN = np.copy( dd[:,:N] )
                
                ## mean in [z], leave [t]
                ddN = np.mean( ddN , axis=0 )
                
                ## assert shape
                if ( ddN.shape != (N,) ):
                    raise ValueError
                    print(f'rank {self.rank:d}: shape violation')
                    self.comm.Abort(1)
                
                d_mean = np.mean( ddN , dtype=np.float64 ) ## [t] mean, was already avged over [z]
                #dI     = np.copy( ddN - d_mean )
                
                data_mean[scalar][iii,iN] = d_mean
                
                try:
                    #Nsig2_ , S1ovS0_ = calc_var_bmbc( dI  , M=M )
                    Nsig2_ , S1ovS0_ = calc_var_bmbc( ddN , M=M )
                except Exception as ee:
                    print(f"error occurred on rank {self.rank:d}: {ee}")
                    traceback.print_exc()
                    #print('exception_traceback : \n'+traceback.format_exc().rstrip())
                    self.comm.Abort(1)
                
                data_Nsig2[scalar][iii,iN] = Nsig2_
                data_ci[scalar][iii,iN,:] = confidence_interval_unbiased(mean=d_mean, N_sigma2=Nsig2_, N=N, confidence=confidence )
            
            #self.comm.Barrier()
            if verbose: progress_bar.update()
            #break ## debug
        
        self.comm.Barrier() ## per scalar Barrier
    
    self.comm.Barrier() ## full loop Barrier
    if verbose:
        progress_bar.close()
        print(72*'-')
    
    # ==============================================================
    # write HDF5 (.h5) file
    # ==============================================================
    
    ## open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## write numpy arrays
            hfw.create_dataset( 'dims/x' , data=x ) ## [m]
            hfw.create_dataset( 'dims/z' , data=z ) ## [m]
            hfw.create_dataset( 'dims/t' , data=t ) ## [s]
            
            ## initialize datasets
            for scalar in scalars:
                hfw.create_dataset( f'mean/{scalar}'  , shape=(ni,nN,)   , dtype=np.float64, chunks=( min(nir,64),nN,  ) )
                hfw.create_dataset( f'Nsig2/{scalar}' , shape=(ni,nN,)   , dtype=np.float64, chunks=( min(nir,64),nN,  ) )
                hfw.create_dataset( f'ci/{scalar}'    , shape=(ni,nN,2,) , dtype=np.float64, chunks=( min(nir,64),nN,2 ) )
            
            ## independently write data
            hfw.create_dataset( 'Ks' , data=Ks )
            hfw.create_dataset( 'Ns' , data=Ns )
    
    self.comm.Barrier()
    
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## collectively write data
        for scalar in scalars:
            
            dset = hfw[f'mean/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = data_mean[scalar][:,:]
            dset = hfw[f'Nsig2/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = data_Nsig2[scalar][:,:]
            dset = hfw[f'ci/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:,:] = data_ci[scalar][:,:,:]
    
    ## report file size
    if verbose:
        even_print( os.path.basename(fn_h5_out), f'{(os.path.getsize(fn_h5_out)/1024**2):0.2f} [MB]' )
        print(72*'-')
    
    ## report file contents
    self.comm.Barrier()
    if (self.rank==0):
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : spd.calc_mean_uncertainty_BMBC() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import numpy as np
from scipy.integrate import solve_ivp
from scipy.optimize import fsolve

# ======================================================================


def Blasius_solution(eta,c0=0.3320573362149271):
    '''
    Blasius similarity ODE:
        f‴ + (1/2)·f·f″ = 0
    
    BCs:
        f(0)=0,  f′(0)=0,  f′(∞)=1
    
    State:
        y=[f, f′, f″],  y′=[f′, f″, -(1/2)·f·f″]
    
    Returns:
        f(η), f′(η)=u/U∞, f″(η)
    '''
    
    scalar_input = np.isscalar(eta)
    eta = np.atleast_1d(np.asarray(eta, dtype=float))
    
    if eta.size > 1 and not np.all(np.diff(eta) > 0):
        raise ValueError("eta must be strictly increasing")
    
    def Blasius_rhs(_,y): ## _ is dummy 't' for solve_ivp
        f, fp, fpp = y
        return [fp, fpp, -0.5 * f * fpp]
    
    ## Shooting for f′′(0)
    if False:
        
        def residual(c0):
            sol = solve_ivp(
                Blasius_rhs,
                (0., eta_test[-1]),
                y0=[0., 0., float(c0)],
                method='RK45',
                atol=1e-12,
                rtol=1e-12
                )
            return 1. - sol.y[1, -1]
        
        eta_test = np.linspace(0., 20., int(1e4))
        
        c0 = fsolve(
                residual,
                x0=0.33205733621490,
                xtol=1e-12
                )[0]
        print(f'c0={c0:0.16f}')
    
    ## Integration
    sol = solve_ivp(
            Blasius_rhs,
            (0., eta[-1]),
            y0=[0., 0., c0],
            method='RK45',
            atol=1e-12,
            rtol=1e-12,
            dense_output=True
            )
    
    if not sol.success:
        raise RuntimeError(sol.message)
    
    f, fp, fpp = sol.sol(eta)
    
    if scalar_input:
        return f[0], fp[0], fpp[0]
    
    return f, fp, fpp
